/*  Copyright Spencer Hakim 2010. All rights reserved.
 *  License: http://www.gnu.org/licenses/gpl-3.0.txt
 *  This is a threaded HTTP server. Errors are logged to appropriate file, as defined by syslog.conf
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <pthread.h>
#include <regex.h>

//Quality assurement and debugging macros
#define QA(x) if(!(x)){syslog(LOG_ERR, "Error encountered at line %d in file %s\n", __LINE__, __FILE__); return 1;}
#define QAP(x, format, ...) if(!(x)){syslog(LOG_ERR, "%s:%d", __FILE__, __LINE__); syslog(LOG_ERR, format, __VA_ARGS__); return 1;}


//global constants and variables
#define BUFSIZE 8196 //based off Apache's LimitLineRequest directive
#define THREADCOUNT 5
#define MAXFD 64

int sock;
struct sockaddr_in server;
FILE* weblog;

typedef struct
{
	int sock;
	struct sockaddr_in from
} connection;

int daemon_init(const char *pname, int facility)
{
	int i;
	
	//Makes it easier for GDB, since it doesn't follow forks by default
#ifndef DEBUG
	pid_t pid = fork();
	
	//catch forking error
	QAP(pid >= 0, "Error forking.\n", NULL);
	
	//parent process terminates 
	if( pid > 0 )
		exit(0);
	
	//sets a new session id, so that shell cannot send a kill signal
	QAP(setsid() >= 0, "Error on setsid().\n", NULL);
#endif
	
	signal(SIGHUP, SIG_IGN); //ignore the hangup signal
	
#ifndef DEBUG
	pid = fork();
	
	//this guarantees that the child is not a session leader
	//and so it cannot obtain a controlling terminal
	if (pid > 0)
		exit(0);
#endif
	
	//This guarantees that anything written to stdout
	//or stderr will not cause a seg fault.
	for(i=0; i < MAXFD; i++)
		close(i);
	open("/dev/null", O_RDONLY);
	open("/dev/null", O_RDWR);
	open("/dev/null", O_RDWR);
	
	//open the log file
	openlog(pname, LOG_PID, facility);
	return 0;
}

void* ThreadRoutine(void* arg)
{
	connection con = *((connection*)arg);
	char buffer[BUFSIZE] = "\0";
	char retMsg[BUFSIZE] = "\0";
	int n, found;
	regex_t getFile;
	regmatch_t matches[2];
	char filename[BUFSIZE] = "\0";
	FILE* temp;
	struct stat st;
	char tempStr[BUFSIZE] = "\0";
	
	//ensure that request starts with / and not ..
	QA( regcomp(&getFile, "^GET /(.+) HTTP/1\\.1", REG_EXTENDED | REG_ICASE) == 0 );
	
	//receive data
	n = recv(con.sock, buffer, BUFSIZE-1, 0);
	QAP(n >= 0, "Error reading from socket.\n", NULL);
	buffer[n] = '\0'; //make sure 0 is set at the end
	
	//write to weblog
	fputs("REMOTE ADDRESS: ", weblog);
	fputs(inet_ntoa(con.from.sin_addr), weblog);
	fputs("\n", weblog);
	fputs(buffer, weblog);
	fflush(weblog);
	
	//get file info
	regexec(&getFile, buffer, 2, matches, 0);
	strncpy(filename, buffer + matches[1].rm_so, matches[1].rm_eo - matches[1].rm_so);
	
	//handle "GET / HTTP/1.1"
	if( strcmp(filename, "") == 0 )
	{
		sprintf(tempStr, "ls %s > .temp", filename);
		system(tempStr);
		strcpy(filename, ".temp");
	}
	
	//get file info
	found = stat(filename, &st);
	
	//build return header
	strcpy(retMsg, "HTTP/1.1 ");
	
	if( found == -1 )
	{
		strncat(retMsg, "404 NOT FOUND\r\n", BUFSIZE - strlen(retMsg));
		strncat(retMsg, "Status: 404 NOT FOUND\r\n", BUFSIZE - strlen(retMsg));
		strncat(retMsg, "\r\n", BUFSIZE - strlen(retMsg));
	}
	else
	{
		strncat(retMsg, "200 OK\r\n", BUFSIZE - strlen(retMsg));
		strncat(retMsg, "Status: 200 OK\r\n", BUFSIZE - strlen(retMsg));
		
		strncat(retMsg, "Content-length: ", BUFSIZE - strlen(retMsg));
		sprintf(retMsg, "%s%u\r\n", retMsg, st.st_size);
		
		strncat(retMsg, "Content-type: ", BUFSIZE - strlen(retMsg));
		sprintf(retMsg, "%s%s\r\n", retMsg, "text/html");
		
		strncat(retMsg, "\r\n", BUFSIZE - strlen(retMsg));
	}
	
	//send headers
	n = send(con.sock, retMsg, strlen(retMsg), 0);
	QAP(n == strlen(retMsg), "Error sending headers.\n", NULL);
	
	//send body
	if( found == 0 )
	{
		temp = fopen(filename, "r");
		
		//send chunks at a time
		while( !feof(temp) )
		{
			memset(retMsg, 0, BUFSIZE);
			fread(retMsg, 1, BUFSIZE, temp);
			n = send(con.sock, retMsg, strlen(retMsg), 0);
			QAP(n == strlen(retMsg), "Error sending body.\n", NULL);
		}
		
		fclose(temp);
	}
	else
	{
		//print a text body for the user
		strcpy(retMsg, "404 File not found");
		n = send(con.sock, retMsg, strlen(retMsg), 0);
		QAP(n == strlen(retMsg), "Error sending body.\n", NULL);
	}
	
	//delete .temp file
	if( strcmp(filename, ".temp") )
	{
		system("rm .temp");
	}
	
	//close the connection
	close(con.sock);
	regfree(&getFile);
	free(arg);
	return 0;
}

int main(int argc, char *argv[])
{
	int retVal;
	pthread_t threadId;
	
	//check for arguments
	if( argc < 2 )
	{
		fprintf(stderr, "Usage: %s [port]\n", argv[0]);
		exit(1);
	}
	
	//spawn daemon
	daemon_init("hakims_HW7_server", LOG_DAEMON);
	
	weblog = fopen("weblog", "a");
	
	//open socket
	sock = socket(AF_INET, SOCK_STREAM, 0);
	QAP(sock >= 0, "Error opening socket.\n", NULL);
	
	//nullify server, and initialize
	memset( (char*)&server, 0, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
	server.sin_port = htons( (unsigned short)atoi(argv[1]) );
	
	//bind socket
	if( bind(sock, (struct sockaddr*)&server, sizeof(server)) < 0 )
		QAP(0, "Error binding socket.\n", NULL);
	
	//listen for incoming connections
	if( listen(sock, THREADCOUNT) < 0 ) //backlog set to # of threads
		QAP(0, "Error listening to socket.\n", NULL);
	
	//keep accepting connections
	while(1)
	{
		//sleeps until connection to accept is available
		connection* con = malloc( sizeof(connection) );
		socklen_t fromSize = sizeof(con->from);
		
		//accept sometimes returns a bad FD, but looping until a good one is returned works.
		do
		{
			con->sock = accept(sock, (struct sockaddr*)&(con->from), &fromSize);
		}
		while(con->sock < 0);
		
		
		//spawn thread
		retVal = pthread_create(&threadId, NULL, &ThreadRoutine, con);
		QAP(retVal == 0, "Error in pthread_create.\n", NULL);
	}
	
	return 0;
}
